# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.

import warnings

warnings.warn(
    "`megatron.core.inference.modelopt_support` will be deprecated in a"
    "future release. Use `megatron.core.post_training.modelopt` instead."
)
